###############################################################################
# Auto-generated by `jupyter-book config`
# If you wish to continue using _config.yml, make edits to that file and
# re-generate this one.
###############################################################################
author = "Helical Team"
autosummary_generate = True
bibtex_bibfiles = ["references.bib"]
comments_config = {"hypothesis": False, "utterances": False}
copyright = "2024"
exclude_patterns = ["**.ipynb_checkpoints", ".DS_Store", "Thumbs.db", "_build"]
extensions = [
    "sphinx_togglebutton",
    "sphinx_copybutton",
    "myst_nb",
    "jupyter_book",
    "sphinx_thebe",
    "sphinx_comments",
    "sphinx_external_toc",
    "sphinx.ext.intersphinx",
    "sphinx_design",
    "sphinx_book_theme",
    "sphinx_inline_tabs",
    "sphinx_proof",
    "sphinx_examples",
    "hoverxref.extension",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinxcontrib.bibtex",
    "sphinx_jupyterbook_latex",
    "sphinx_multitoc_numbering",
]
external_toc_exclude_missing = False
external_toc_path = "_toc.yml"
hoverxref_intersphinx = ["sphinxproof"]
html_baseurl = ""
html_css_files = ["my_theme copy.css", "my_theme.css"]
html_favicon = "docs/official_logo_fav.png"
html_logo = "docs/official_logo_small.png"
html_sourcelink_suffix = ""
html_static_path = ["_static"]
html_theme = "sphinx_book_theme"
html_theme_options = {
    "search_bar_text": "Search this book...",
    "launch_buttons": {
        "notebook_interface": "classic",
        "binderhub_url": "",
        "jupyterhub_url": "",
        "thebe": False,
        "colab_url": "",
    },
    "path_to_docs": "docs",
    "repository_url": "https://github.com/helicalAI/helical",
    "repository_branch": "main",
    "extra_footer": "",
    "home_page_in_toc": True,
    "announcement": "",
    "analytics": {"google_analytics_id": ""},
    "use_repository_button": True,
    "use_edit_page_button": False,
    "use_issues_button": True,
}
html_title = "Helical Documentation"
intersphinx_mapping = {
    "ebp": ["https://executablebooks.org/en/latest/", None],
    "myst-parser": ["https://myst-parser.readthedocs.io/en/latest/", None],
    "myst-nb": ["https://myst-nb.readthedocs.io/en/latest/", None],
    "sphinx": ["https://www.sphinx-doc.org/en/master", None],
    "nbformat": ["https://nbformat.readthedocs.io/en/latest", None],
    "sd": ["https://sphinx-design.readthedocs.io/en/latest", None],
    "sphinxproof": ["https://sphinx-proof.readthedocs.io/en/latest/", None],
}
latex_engine = "pdflatex"
mathjax3_config = {
    "tex": {
        "macros": {
            "N": "\\mathbb{N}",
            "floor": ["\\lfloor#1\\rfloor", 1],
            "bmat": ["\\left[\\begin{array}"],
            "emat": ["\\end{array}\\right]"],
        }
    }
}
myst_enable_extensions = [
    "colon_fence",
    "dollarmath",
    "linkify",
    "substitution",
    "tasklist",
]
myst_url_schemes = ["mailto", "http", "https"]
napoleon_include_init_with_doc = True
napoleon_numpy_docstring = True
nb_execution_allow_errors = False
nb_execution_cache_path = ""
nb_execution_excludepatterns = []
nb_execution_in_temp = False
nb_execution_mode = "off"
nb_execution_timeout = 30
nb_output_stderr = "show"
numfig = True
pygments_style = "sphinx"
suppress_warnings = ["myst.domains"]
use_jupyterbook_latex = True
use_multitoc_numbering = True
